package prowas3

import groovy.transform.Immutable


// groovy First.groovy [folder]


final int START_NUMBER = 63


def input = args.size() ? args[0] : "vzorova_zadani"

def files = []

new File(input).eachFile {
	 files << it
}

files = files.grep{ it.isFile() && it.name.endsWith(".in") }.sort { it.name }

files.each {
	Input i = parseInput(it.text)
	new File(it.parent, it.name[0..-3] + START_NUMBER)
		.write(makeSolution(i))
}


Input parseInput(String str) {
	def ret = new Input()
	def lines = str.split("\n")
	lines.first().split(/[\s]+/).with {
		ret.basew = it[0] as int
		ret.baseh = it[1] as int
		switch(it[2]) {
		case "P":
			ret.type = Type.P
		break
		case "D":
			ret.type = Type.D
		break
		case "PB":
			ret.type = Type.PB
		break
		case "DB":
			ret.type = Type.DB
		break
		default:
			println "Unknown type!"
		}
	}
	lines[1..-1].each {
		def split = it.split(/[\sX\n]+/)
		def cube = new Cube(
			color: split[0],
			width: split[1] as int,
			height: split[2] as int)
		ret.cubes << cube
		ret.ccount[cube] = split[3] as int
	}
	return ret
}

String makeSolution(Input inp) {
	println "-------------------- ${inp.type}"
	def ret = "${inp.baseh} ${inp.basew} ${inp.type}\n" << ""
	inp.cubes = inp.cubes.sort { -(it.width + it.height) } // Prefer bigger cubes
	Floor prev
	for(int flc = 1;;flc++) {
		Floor flr = new Floor(width: inp.basew, 
				height: inp.baseh, type: inp.type, under: prev)
		println flr.allEmptyCoords()
		inp.cubes.each { c ->
			if(inp.ccount[c] <= 0)
				return // We could remove, but that just lowers performance (and throws ConcurrentModificationException for some reason)
			flr.allEmptyCoords().each {
				if(inp.ccount[c] <= 0)
					return
				// I think GC is going to love this...
				def placed = new PlacedCube(x: it.x, y: it.y, 
						width: c.width, height: c.height, color: c.color)
				if(flr.doesFit(placed)) {
					ret << "${flc} ${it.x + 1} ${it.y + 1} ${c.color} ${placed.width}x${placed.height}\n"
					flr.cubes << placed
					inp.ccount[c]--
				}
				else { // Rotate
					int tmp = placed.width
					placed.width = placed.height
					placed.height = tmp 
					// This really needs to be fixed
					if(flr.doesFit(placed)) {
						ret << "${flc} ${it.x + 1} ${it.y + 1} ${c.color} ${placed.width}x${placed.height}\n"
						flr.cubes << placed
						inp.ccount[c]--
					}
				}
			}
		}
		println "Remaining:"
		inp.cubes.each {
			println "${it.color} : ${it.width}x${it.height} -- ${inp.ccount[it]}"
		}
		if(!flr.isCompleted()) {
			println "Failed ${flr.allEmptyCoords()}"
			break
		}
		else {
			println "Completed floor ${flc}"
		}
		prev = flr
	}
	return ret
}

class Floor {
	int width
	int height
	Floor under
	Type type
	List<PlacedCube> cubes = []
	
	boolean doesFit(PlacedCube c) {
		if(c.allCoords().any { isFull(it) })
			return false
		if((type == Type.D || type == Type.DB) && under)
			if(!c.allCoords().any { under.isFull(it) })
				return false
		if(type == Type.DB || type == Type.PB) {
			def bor = c.allCoords()
				.grep { it.x == width - 1 || it.y == height - 1 || it.x == 0 || it.y == 0 }
			if(under && bor.any {
					def u = under.getAt(it)
					u && u.color == c.color
				}) {
				return false
			}
			def neigh = []
			bor.each {
				int x
				int y
				if(it.x == 0 || it.x == width - 1) {
					neigh << new Coord(it.x, it.y - 1)
					neigh << new Coord(it.x, it.y + 1)
				}
				else if(it.y == 0 || it.y == height - 1) {
					neigh << new Coord(it.x + 1, it.y)
					neigh << new Coord(it.x - 1, it.y)
				}
			}
			if(neigh.any {
				def n = getAt(it)
				if(n && n.color == c.color)
					return true
			})
				return false
		}
		return true
	}
	
	PlacedCube getAt(Coord c) {
		if(c.x >= width || c.y >= height || c.x < 0 || c.y < 0)
			return null // Out of bounds are automatically considered full
		for(it in cubes) {
			if(it.allCoords().grep {
					it.x == c.x && it.y == c.y
				})
			return it
				
		}
		return null
	}
	
	boolean isFull(Coord c) {
		if(c.x >= width || c.y >= height || c.x < 0 || c.y < 0)
			return true // Out of bounds are automatically considered full
		cubes.any { // Efficiency is for cowards
			it.allCoords().any {
				it.x == c.x && it.y == c.y
			}
		}
	}
	
	boolean isCompleted() {
		!allEmptyCoords()
	}
	
	def allEmptyCoords() { // This is just stupid
		if(type == Type.P || type == Type.PB) {
			return new PlacedCube(x: 0, y: 0, width: width, height: height)
					.allCoords()
					.grep { !isFull(it)}
		}
		else {
			def ret = [] as Set
			(0..<width).each { w ->
				ret << new Coord(x: w, y: height - 1)
				ret << new Coord(x: w, y: 0)
			}
			(0..<height).each { h ->
				ret << new Coord(x: width - 1, y: h)
				ret << new Coord(x: 0, y: h)
			}
			return ret.grep { !isFull(it) }
		}
	}
}

class Input {
	int basew // X axis
	int baseh // Y axis
	Type type
	def cubes = []
	def ccount = [:]
}

class PlacedCube extends Cube {
	int x
	int y
	
	def allCoords() {
		(x..<(x + width)).collect { w ->
			(y..<(y + height)).collect { h ->
				new Coord(x: w, y: h)
			}
		}.flatten()
	}
}

class Cube {
	String color
	int width // X axis
	int height // Y axis
}

@Immutable
class Coord {
	int x
	int y
}

enum Type {
	P, D, PB, DB
}